# $Id: upgrade.py,v 1.14 2001/03/13 13:33:25 olafb Exp $

#  Copyright 2001 Trustix As - <http://www.trustix.comQ>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

true, false = 1, 0

INSTALL_FLAG = "install"
UPGRADE_FLAG = "upgrade"

DEPENDENCY_ERROR_MESSAGE = \
"""There were some problems with dependencies when upgrading or
installing.  Usually problems arise when an upgrade will break already
installed software. In order to upgrade the selected package(s), you
must also upgrade the already installed (and depending) packages. A
'swup --upgrade' will usually fix this."""

NOSITES_ERROR_MESSAGE = \
"""Unable to open a site or get valid upgrade information. This will
happen if all top priority sites are down, or if upgrade information
is not available or is missing a valid signature from a known public
key. Check site, configuration and public keyring."""


# exceptions
download_error = "upgrade.download_error"
save_to_error = "upgrade.save_to_error"
signature_error = "upgrade.signature_error"
resolve_error = "upgrade.resolve_error"
signature_invalid = "upgrade.signature_invalid"
install_error = "upgrade.install_error"
flush_cache_error = "upgrade.flush_cache_error"
driver_error = "upgrade.driver_error"
parse_error = "upgrade.parse_error"

import sys, re, os, string
import spiparser, download, log
import resolver
import utils


is_uri_regexp = re.compile( "(.rdf)|(^http:)|(^ftp:)" )



def is_uri( uri ):
    if is_uri_regexp.search( uri ):
        return true
    else:
        return false


def _sort_2d_sitelist(element1, element2):
    pri1, pri2 = abs( int(element1[0]) ), abs( int(element2[0]) )
    if  pri1 < pri2:
        return -1
    elif pri1 > pri2:
        return 1
    elif pri1 == pri2:
        return 0
    else:
        raise ValueError, "unable to compare"
    

def _compile_regexps( regexps ):
    """Compile list of strings into regexp objects. Returns a list of
    re.RegexObject instances."""
    retlist = []
    for regexp in regexps:
        if regexp == None:
            retlist.append( re.compile(".(?!.)") )
        else:
            retlist.append( re.compile(regexp) )
    return retlist




class upgrade:
    "General class for upgrade/install. To be subclassed."

    def _load_driver( self, drivername ):
        try:
            exec "import %s; self.pkgdrv = %s.driver(self.log)" %\
                 (drivername, drivername)
        except ImportError:
            message = "Unable to load driver for packaging system. - "+\
                      "Exititing."
            self.log.write_log( message )
            self.log.write_stderr( message )
            self.log.write_syslog_warning( message )
            sys.exit( 1 )


    def __init__( self, configdict=None ):
        if not configdict:
            import config
            self.config = config.parse()
        else:
            self.config = configdict

        if string.lower( self.config["interactive_logged"] ) == "yes":
            interactive_logged = true
        else:
            interactive_logged = false

        self.log = log.log( self.config["logfile"], self.config["loglevel"],
                            interactive_logged )
        self.reader = spiparser.rdf_reader()
        
        drivername = self.config["pkgdriver"]
        self._load_driver( drivername )
        self._time = 0
        self._downloader = download.download( self.log,
                                              self.config["cachedir"],
                                              self.config["tmpdir"],
                                              self.config["md5sumfile"])
            
        regexps = _compile_regexps( [self.config["exclude_pkg_regexp"],
                                     self.config["include_pkg_regexp"],
                                     self.config["exclude_group_regexp"],
                                     self.config["include_group_regexp"]] )
        (self.exclude_pkg_regexp,
         self.include_pkg_regexp,
         self.exclude_group_regexp,
         self.include_group_regexp) = regexps


    def load_data(self, filename):
        import pickle
        fd = open(filename)
        retval = pickle.load(fd)
        fd.close()
        return retval


    def save_data(self, obj, filename):
        import pickle
        dirname = os.path.dirname(filename)
        if not os.path.isdir(dirname):
            os.makedirs(dirname)
        fd = open(filename, 'w')
        pickle.dump(obj, fd)
        fd.close()

        
    def get_latest(self, uri):
        """Parse package list rdf, given the uri to the rdf-file.  Returns
        a dictionary with package name as key, and the a dictionary of
        name, version, release, summary, uri and obsoletes as value for
        keys."""
        # Check md5sum, pickle.load and/or dump -- to be implemented.

        # Get the md5sum uri:
        if uri[-3:] == '.gz':
            md5sum_uri = uri[:-3] + '.md5'
            noncompressed_uri = uri[:-3]
        else:
            noncompressed_uri = uri
            md5sum_uri = uri + '.md5'

        dict = None

        # Check if md5sum has changed. The check will update md5sums
        # if they are changed.
        if self.md5sum_isequal(md5sum_uri, noncompressed_uri):
            # Pickle already parsed latest-information.
            pickle_file = os.path.join(self.config['listsdir'], uri)
            pickle_file = utils.normalize_uri(pickle_file + '.pickle')
            if os.path.isfile(pickle_file):
                dict = self.load_data(pickle_file)
                
        if not dict:
            # Get latest, parse, and pickle-dump.
            local_filename = self.download_file(uri)
            (baseuri, basename) = os.path.split(uri)
            self.log.write_tty('\nParsing file: %s -' % basename)
            dict = self.reader.read_latest( local_filename, baseuri )
            self.log.write_tty('\rParsing file: %s - done\n' % basename)
            pickle_file = os.path.join(self.config['listsdir'], uri)
            pickle_file = utils.normalize_uri(pickle_file + '.pickle')
            self.save_data(dict, pickle_file)
        return dict



    def get_local_info( self ):            
        installed_pkgs = self.pkgdrv.get_installed_pkgs()
        return installed_pkgs


    def get_info_by_uris( self, pkg_uri ):
        pkg_list = []
        
        for uri in pkg_uri:
            try:
                (baseuri, basename) = os.path.split( uri )
                local_filename = self.download_file( uri )
                dict = self.reader.read_package_short( local_filename,
                                                       baseuri )
                if dict:
                    pkg_list.append( dict )
            except spiparser.parse_error:
                raise parse_error, "Unable to parse file '%s'" % uri
        return pkg_list


    def get_info_by_names( self, pkg_name_list ):
        sitelist = self.config["sites"]
        sitelist.sort( _sort_2d_sitelist )

        pkg_uri = []
        pri_handled = []
        upgrade_uris = []

        next_pri = 0
        for site in sitelist:
            if len ( pkg_name_list ) == len ( pkg_uri ): break
            (pri, latest_uri, regexp) = site
            pri = abs( int(pri) )
            if pri in pri_handled: continue
            elif pri > next_pri: break

            try:
                message = "Fetching upgrade info '%s'" % latest_uri
                self.log.write_log( message )
                self.log.write_stdout( '\r' + message )
                latest = self.get_latest( latest_uri )
                who_obsoletes = self.find_who_obsoletes( latest )
                
            except (download_error, parse_error), errmsg:
                self.log.write_log(errmsg)
                self.log.write_stderr(errmsg)
                continue

            pri_handled.append( pri )
            for pkg_name in pkg_name_list:
                if latest.has_key( pkg_name ) and \
                   re.compile( regexp ).match( pkg_name ):
                    pkg_uri.append( latest[pkg_name] )
                elif who_obsoletes.has_key( pkg_name ) and \
                     re.compile( regexp ).match( pkg_name ):
                    # Just pick the first package found that obsoletes.
                    new_pkg_name = who_obsoletes[pkg_name][0]
                    pkg_uri.append( latest[new_pkg_name] )
                    message = "'%(old)s' is obsoleted by '%(new)s'."\
                              " Will use '%(new)s'." % {"old": pkg_name,
                                                       "new": new_pkg_name}
                    self.log.write_log( message )
                    self.log.write_stdout( message )
                else:
                    pkg_uri.append( None )

            next_pri = pri + 1
        return pkg_uri


    def find_who_obsoletes( self, latestdict ):
        """Given a dictionary over latest packages, return a dictionary
        with obsoleted package as key and a list of packages that obsoletes
        the package as value for key."""
        
        obsoleted_by = {}
        for name in latestdict.keys():
            if latestdict[name]["obsoletes"]:
                for obs in latestdict[name]["obsoletes"]:
                    if not obsoleted_by.has_key( obs ):
                        obsoleted_by[obs] = [name]
                    else:
                        obsoleted_by[obs].append( name )
        return obsoleted_by
        

    def regexp_filter( self, name, group ):
        """Filter (<name>, <group>) through regexps given in configuration
        file. Returns 1 if package is to be upgraded or installed, 0
        otherwise."""
        if self.exclude_pkg_regexp.match( name ):
            return false
        
        if not self.include_pkg_regexp.match( name ):
            return false
        
        if self.exclude_group_regexp.match( group ):
            return false

        if not self.include_group_regexp.match( group ):
            return false

        return true
        

    def config_regexps_filter( self, package_dict ):
        upgradelist = []
        for name in package_dict.keys():
            if len( package_dict[name] ) > 1:
                message ='Multiple versions installed.'\
                          ' Skipping package %s' % name
                self.log.write_stderr( message )
                self.log.write_log( message )
                continue
            
            for package in package_dict[name]:
                (version, release, group) =  package
                if self.regexp_filter( name, group ):
                    upgradelist.append( (name, version, release, group) )
        return upgradelist


    def md5sum_isequal(self, md5sumuri, uri):
        return self._downloader.md5sumcheck(md5sumuri, uri)

        
    def download_file( self, uri, siguri=None ):
        "Download file from uri and check signature. Returns local filename."
        try:
            filename = self._downloader.download( uri, siguri )
            return filename
        except self._downloader.error, errmsg:
            raise download_error, errmsg
        

    def download( self, pkglist ):
        for i in range( len(pkglist) ):
            uri = utils.normalize_uri( pkglist[i][1] )
            siguri = utils.normalize_uri( pkglist[i][3] )
            pkglist[i][1] = self.download_file( uri, siguri )
        return pkglist


    def resolve_deps( self, pkglist ):
        dep_resolver = resolver.resolver( self.reader, self.regexp_filter,
                                          self.pkgdrv, self.download_file,
                                          self.log )
        try:
            pkglists = dep_resolver.solve_deps( pkglist )
            pkglists_info = "" 
            for i in range( len(pkglists) ):
                pkglist = pkglists[i]
                for pkg in pkglist:
                    pkglists_info = pkglists_info + \
                                   "\n\t%s: %s, version %s release %s" % \
                                   (pkg[2], pkg[0], pkg[4], pkg[5])
            if pkglists_info != "":
                message = "Packages to install/upgrade:%s" % pkglists_info
                self.log.write_log(message)
                self.log.write_stdout(message)
        except (resolver.DependencyError, resolver.ParseError), errmsg:
            raise resolve_error, errmsg

        return pkglists


    def flush_cache( self ):
        if string.lower( str(self.config["flush_cache"]) ) == "yes":
            try:
                self._downloader.flush_cache()
            except download.flush_cache_error, msg:
                raise flush_cache_error, msg


    def install( self, pkglist ):
        try:
            self.pkgdrv.install( pkglist )
            return None
        except self.pkgdrv.install_error, errmsg:
            return errmsg


    def _run( self, pkglist ):
        if string.lower( str(self.config["poll_only"]) ) == "yes":
            message = "Found these packages:\n"
            for pkg in pkglist:
                message = message + "\t%s: %s-%s-%s\n" %\
                          (pkg[2], pkg[0], pkg[4], pkg[5])
            message = message + "\tTotal %i packages." % len( pkglist )
            self.log.write_stdout( message )
            self.log.write_log( message )
        else:
            pkglists = self.resolve_deps( pkglist )
            if not pkglists:
                message = "Nothing to install or upgrade."
                self.log.write_stdout( message )
                self.log.write_log( message )
                return
            i = 0
            errorflag = false

            for pkglist in pkglists:
                try:
                    downloadlist = self.download( pkglist )
                    message = "Download completed for transaction %i." % i
                    self.log.write_stdout_verbose( message )
                    
                except download_error, errmsg:
                    skipped = ""
                    for package in pkglist:
                        skipped = skipped + package[0] + " "
                    skipped = string.rstrip( skipped )
                    errmsg = errmsg + "\nSkipping packages %s" % skipped
                    self.log.write_log( errmsg )
                    self.log.write_stderr( errmsg )
                    i = i +1
                    continue

                if self.config['save_to']:
                    todir = self.config['save_to']
                    import shutil
                    if not os.path.isdir( todir ):
                        message = 'Can not save copy of packages.'\
                                  ' No such directory'\
                                  ': %s' % todir
                        self.log.write_log( message )
                        self.log.write_stderr( message )
                    else:
                        for package in downloadlist:
                            try:
                                fromfile = package[1]
                                basename = os.path.basename( fromfile )
                                tofile = os.path.join( todir, basename )
                                shutil.copyfile( fromfile, tofile )
                            except IOError:
                                message = 'Unable to move file %s to %s' % \
                                          (fromfile, tofile)
                                self.log.write_stderr( message )
                                self.log.write_log( message )
          
                if not string.lower( str(self.config["download_only"]) ) == \
                   "yes":
                    errmsg = self.install( downloadlist )
                    if errmsg:
                        errorflag = true
                        self.log.write_log( errmsg )
                        self.log.write_stderr( errmsg )
                    else:
                        self.flush_cache()

                i = i +1
            if errorflag:
                self.log.write_stderr( DEPENDENCY_ERROR_MESSAGE )
                

class upgrade_normal( upgrade ):
    def __init__( self, configdict = None ):
        upgrade.__init__( self, configdict )


    def select_packages( self ):
        installed_pkgs = self.pkgdrv.get_installed_pkgs()
        upgradelist = self.config_regexps_filter( installed_pkgs )
        if not upgradelist: return []
        nupgradelist = len( upgradelist )

        sitelist = self.config["sites"]
        sitelist.sort( _sort_2d_sitelist )
        
        pri_handled = []
        return_list = []
        next_pri = 0
        for site in sitelist:
            (pri, latest_uri, regexp) = site
            pri = abs( int(pri) )
            if pri in pri_handled: continue
            elif pri > next_pri: break
            elif not regexp:
                pri_handled.append( pri )
                continue
            try:
                message = "Fetching upgrade info '%s'" % latest_uri
                self.log.write_log( message )
                self.log.write_stdout( '\r' + message )
                latest = self.get_latest( latest_uri )
            except (download_error, parse_error), errmsg:
                errmsg = errmsg + " Skipping."
                self.log.write_log( errmsg )
                self.log.write_stderr( errmsg ) 
                continue
            pri_handled.append( pri )
            next_pri = pri + 1
            i = 0

            obsoleted_by = self.find_who_obsoletes( latest )
                
            while i < nupgradelist:
                name = upgradelist[i][0]
                if latest.has_key( name ) and re.match( regexp, name ):
                    latest_version = latest[name]["version"]
                    latest_release = latest[name]["release"]
                    latest_summary = latest[name]["summary"]
                    latest_uri = latest[name]["uri"]
                    current_version, current_release = upgradelist[i][1:3]
                    if utils.version_cmp( (latest_version, latest_release),
                                   (current_version, current_release) ) == 1:
                        return_list.append( [name, latest_uri, UPGRADE_FLAG,
                                             None, latest_version,
                                             latest_release] )
                        del( upgradelist[i] )
                        nupgradelist = nupgradelist -1
                    else:
                        i = i + 1

                elif obsoleted_by.has_key( name ):
                    # We just pick the first package that obsoletes another.
                    oldname = name
                    name = obsoleted_by[name][0]
                    latest_version = latest[name]["version"]
                    latest_release = latest[name]["release"]
                    latest_summary = latest[name]["summary"]
                    latest_uri = latest[name]["uri"]
                    message = "%s is obsoleted. Will upgrade with %s." % \
                              (oldname, name)
                    self.log.write_stdout( message )
                    self.log.write_log( message )
                    return_list.append( [name, latest_uri,
                                         UPGRADE_FLAG, None,
                                         latest_version,
                                         latest_release] )
                    del( upgradelist[i] )
                    nupgradelist = nupgradelist -1
                    
                else:
                    i = i + 1
            
        if not pri_handled:
            return None
        else: 
            return return_list



    def run( self ):
        message = "Starting upgrade."
        self.log.write_log( message )
        self.log.write_stdout_verbose( message )
        try:
            upgradelist = self.select_packages()
            if upgradelist == None:
                self.log.write_log( NOSITES_ERROR_MESSAGE )
                self.log.write_stderr( NOSITES_ERROR_MESSAGE )
            elif not upgradelist:
                message = "Your system is up to date."
                self.log.write_log( message )
                self.log.write_stdout( message )
            else:
                self._run( upgradelist )
        except ( download_error, resolve_error, install_error,
                 save_to_error ), message:
            self.log.write_log( message )
            self.log.write_stderr( message )
            self.flush_cache()
            sys.exit( 1 )
        except KeyboardInterrupt:
            message = "User interrupt - exiting."
            self.log.write_log( message )
            self.log.write_stderr( "\n"+message )
            self.flush_cache()
            sys.exit( 1 )
        except:
            errmsg = sys.exc_info()[1]
            message = "Fatal error: %s - exiting" % errmsg
            self.log.write_log( message )
            self.log.write_stderr( message )
            self.log.write_syslog_warning( message )
            self.flush_cache()
            sys.exit( 1 )
        self.flush_cache()

        
        

class upgrade_package( upgrade ):
    def __init__( self, install_list, configdict = None ):
        upgrade.__init__( self, configdict )
        self.install_list = install_list


    def is_installed ( self, name ):
        if not self.local_pkgs.has_key( name ):
            message = "Package not installed, will not upgrade '%s'."\
                      % name
            self.log.write_log( message )
            self.log.write_stderr( message )                    
            return None
        else:
            return true
        
        
    def check_package( self, package ):
        if not package:
            return
        name = package["name"]
        uri = package["name"]
        version = package["version"]
        release = package["release"]

        # Check if package is a valid upgrade for an installed package.
        if self.local_pkgs.has_key( name ):
            if len( self.local_pkgs[name] ) > 1:
                message = "Multiple versions installed." \
                          " Skipping package %s" % name
                self.log.write_stderr( message )
                self.log.write_log( message )
                return false
 
            for local_pkg in self.local_pkgs[name]:
                current_version, current_release = local_pkg[:2]
                comparison = utils.version_cmp( (version, release),
                                                (current_version,
                                                current_release) )
                if  comparison > 0:
                    return true
                elif comparison == 0:
                    #Package is already installed, notify the user,
                    #if all packages are already installed exit.
                    message = "Package %s" % name +\
                              " is already at newest version (%s-%s)."\
                              % (current_version, current_release)
                    self.log.write_log( message )
                    self.log.write_stderr( message )
                    return false

        # Check if package obsoletes an installed package, and should
        # therefore be set to upgrade.
        if package.has_key( "obsoletes" ):
            obsoletes = package["obsoletes"]
            for obs_name in obsoletes:
                if self.local_pkgs.has_key( obs_name ):
                    return true

        # Package is not an upgrade or obsoletes an installed package.
        return false
            
            
    def check_list( self ):
        if not self.install_list:
            return None
        
        self.local_pkgs = self.get_local_info()
        resolver_input = []
        siguri = None
        if is_uri( self.install_list[0] ):
            packages = self.get_info_by_uris( self.install_list )

            for i in range( len(packages) ):
                package = packages[i]
                if self.check_package( package ):
                    name = package["name"]
                    uri = self.install_list[i]
                    flag = UPGRADE_FLAG
                    siguri = None
                    version = package["version"]
                    release = package["release"]
                    resolver_input.append( [name, uri, flag, siguri, version,
                                            release] )
            
        else :
            packages = self.get_info_by_names( self.install_list )
	    if not packages:
                self.log.write_log( NOSITES_ERROR_MESSAGE )
                self.log.write_stderr( NOSITES_ERROR_MESSAGE )
            else:
                for i in range( len(self.install_list) ):
                    package = packages[i]
                    name = self.install_list[i]
                    if not package:
                        message = "Nothing found for '%s'." % name
                        self.log.write_log( message )
                        self.log.write_stderr( message )
                        continue

                    if self.check_package( package ):
                        name = package["name"]
                        uri = package["uri"]
                        flag = UPGRADE_FLAG
                        siguri = None
                        version = package["version"]
                        release = package["release"]
                        resolver_input.append( [name, uri, flag, siguri,
                                                version, release] )
        return resolver_input
    

    def run( self ):
        message = "Starting upgrade of package."
        self.log.write_log(message)
        self.log.write_stdout_verbose(message)
        try:
            package_list = self.check_list()
            if package_list:
                self._run( package_list )
        except ( download_error, resolve_error, install_error,
                 save_to_error ), message:
            self.log.write_log( message )
            self.log.write_stderr( message )
            self.flush_cache()
            sys.exit( 1 )
        except KeyboardInterrupt:
            message = "User interrupt - exiting."
            self.log.write_log( message )
            self.log.write_stderr( "\n"+message )
            self.flush_cache()
            sys.exit( 1 )
        except:
            errmsg = sys.exc_info()[1]
            message = "Fatal error: %s - exiting." % errmsg
            self.log.write_log( message )
            self.log.write_stderr( message )
            self.log.write_syslog_warning( message )
            self.flush_cache()
            sys.exit( 1 )

        self.flush_cache()



    
class install_package( upgrade ):
    def __init__( self, install_list, configdict = None ):
        upgrade.__init__( self, configdict )
        self.install_list = install_list


    def check_package( self, package ):
        if not package:
            return
        
        name = package["name"]
        version = package["version"]
        release = package["release"]

        if self.local_pkgs.has_key( name ):
            message = "Already installed package, try upgrading: %s" % name
            self.log.write_stderr( message )
            self.log.write_log( message )
            return false

        # Package is not installed. It is ok to install it.
        return true
        

    def check_list( self ):
        if not self.install_list:
            return None
        
        self.local_pkgs = self.get_local_info()
        resolver_input = []
        siguri = None
        if is_uri( self.install_list[0] ):
            packages = self.get_info_by_uris( self.install_list )
            for  i in range( len(packages) ):
                package = packages[i]
                if self.check_package( package ):
                    name = package["name"]
                    uri = self.install_list[i]
                    flag = INSTALL_FLAG
                    siguri = None
                    version = package["version"]
                    release = package["release"]
                    resolver_input.append( [name, uri, flag, siguri, version,
                                            release] )

        else :
            packages = self.get_info_by_names( self.install_list )
	    if not packages:
                self.log.write_log( NOSITES_ERROR_MESSAGE )
                self.log.write_stderr( NOSITES_ERROR_MESSAGE )
            else:
                for i in range( len(packages) ):
                    package = packages[i]
                    name = self.install_list[i]
                    if not package:
                        message = "Nothing found for '%s'." % name
                        self.log.write_log( message )
                        self.log.write_stderr( message )
                        continue

                    if self.check_package( package ):
                        name = package["name"]
                        uri = package["uri"]
                        flag = INSTALL_FLAG
                        siguri = None
                        version = package["version"]
                        release = package["release"]
                        resolver_input.append( [name, uri, flag, siguri,
                                                version, release] )
        return resolver_input


    # main run function
    def run( self ):
        message = "Starting install."
        self.log.write_log( message )
        self.log.write_stdout_verbose( message )
        try:
            package_list = self.check_list()
            if package_list:
                self._run(package_list)
            else:
                self.flush_cache()
        except ( download_error, resolve_error, install_error,
                 save_to_error ), message:
            self.log.write_log( message )
            self.log.write_stderr( message )
            self.flush_cache()
            sys.exit( 1 )
        except KeyboardInterrupt:
            message = "User interrupt - exiting."
            self.log.write_log( message )
            self.log.write_stderr( message )
            self.flush_cache()
            sys.exit( 1 )
        except:
            errmsg = sys.exc_info()[1]
            message = "Fatal error: %s - exiting." % errmsg
            self.log.write_log( message )
            self.log.write_stderr( message )
            self.log.write_syslog_warning( message )
            self.flush_cache()
            sys.exit( 1 )

        self.flush_cache()




class list_latest( upgrade ):

    def __init__( self, configdict = None ):
        upgrade.__init__(self, configdict)


    def get_data( self ):
        sitelist = self.config["sites"]
        sitelist.sort(_sort_2d_sitelist)
        
        pri_handled = []
        latest = {}
        next_pri = 0
        for site in sitelist:
            (pri, latest_uri, regexp) = site
            pri = abs( int(pri) )
            if pri in pri_handled: continue
            elif pri > next_pri: break
            elif not regexp:
                pri_handled.append( pri )
                continue
            try:
                message = "Fetching info '%s'" % latest_uri
                self.log.write_log( message )
                self.log.write_stderr( '\r' + message )
                dict = self.get_latest( latest_uri )
                for key in dict.keys():
                    if not latest.has_key( key ):
                        latest[key] = dict[key]
            except (download_error, parse_error), errmsg:
                self.log.write_log( errmsg )
                self.log.write_stderr( errmsg )    
                continue
            pri_handled.append( pri )
            next_pri = pri + 1
        if not pri_handled:
            return None
        else: 
            return latest


    def write( self ):
        data = self.get_data()
        if not data: return
        keys = data.keys()
        keys.sort()
        for name in keys:
            version = data[name]["version"]
            release = data[name]["release"]
            uri = data[name]["uri"]
            summary = data[name]["summary"]
            message = "%s-%s-%s - %s" % ( name, version, release, summary )
            self.log.write_stdout( message )

    def run( self ):
        try:
            self.write()
            self.flush_cache()
        except ( download_error, resolve_error, install_error,
                 save_to_error ), message:
            self.log.write_log( message )
            self.log.write_stderr( message )
            self.flush_cache()
            sys.exit( 1 )
        except KeyboardInterrupt:
            message = "User interrupt - exiting."
            self.log.write_log( message )
            self.log.write_stderr( "\n"+message )
            self.flush_cache()
            sys.exit( 1 )
        except:
            errmsg = sys.exc_info()[1]
            message = "Fatal error: %s - exiting." % errmsg
            self.log.write_log( message )
            self.log.write_stderr( message )
            self.log.write_syslog_warning( message )
            self.flush_cache()
            sys.exit( 1 )

        self.flush_cache()


class list_new( list_latest ):
    
    def __init__( self, configdict=None ):
        list_latest.__init__( self, configdict )
        
    def write( self ):
        data = self.get_data()
        if not data: return
        keys = data.keys()
        keys.sort()
        installed_pkgs = self.pkgdrv.get_installed_pkgs()
        for name in keys:
            if self.pkgdrv.packages.has_key( name ): continue
            version = data[name]["version"]
            release = data[name]["release"]
            uri = data[name]["uri"]
            summary = data[name]["summary"]
            message = "%s-%s-%s - %s" % ( name, version, release, summary )
            self.log.write_stdout( message )

